/** \file App.cpp */
#include "App.h"

int main(int argc, char** argv) {
    (void)argc; (void)argv;
    GApp::Settings settings;
    
    settings.window.width       = 960; 
    settings.window.height      = 1080;
    settings.window.caption     = "Contour Stroking Demo - Morgan McGuire";

#   ifdef G3D_WIN32
	    if (FileSystem::exists("data-files", false)) {
            // Running on Windows, building inside the starter directory
            chdir("data-files");
        }
#   endif

    return App(settings).run();
}

// Tells C++ to invoke command-line main() function even on OS X and Win32.
G3D_START_AT_MAIN();

App::App(const GApp::Settings& settings) : 
    GApp(settings), 
    m_style(DEPTH_STYLE),
    m_depthThreshold(0.00175f),
    m_normalThreshold(0.5f) {

#   ifdef G3D_DEBUG
        // Let the debugger catch unhandled exceptions
        catchCommonExceptions = false;
#   endif
}


void App::onInit() {
    makeGBuffer();
    makeLighting();
    makeGui();
    loadShaders();
    loadModel();
}


void App::makeLighting() {
    m_lighting = Lighting::create();
    m_lighting->ambientTop = Color3::white() * 0.25f;
    m_lighting->ambientBottom = Color3::white() * 0.15f;
    m_lighting->shadowedLightArray.append(GLight::directional(Vector3(-1,2,0.5f), Color3::white()));
    m_shadowMap = ShadowMap::create();
}


void App::makeGBuffer() {
    GBuffer::Specification gbufferSpec;
    gbufferSpec.csNormal = true;
    gbufferSpec.normalsAreUnsigned = false;
    gbufferSpec.format = ImageFormat::RGB16F();
    gbufferSpec.depthFormat = ImageFormat::DEPTH32();

    m_gbuffer = GBuffer::create("GBuffer", gbufferSpec);
    m_gbuffer->resize(renderDevice->width(), renderDevice->height());
}


void App::loadModel() {
    // Load the model

    const char* filename = "sphere/sphere.ifs";  const float scale = 3.0f;
    // const char* filename = "ship/VCT-Frigate.3ds";  const float scale = 0.001f;
    
    ArticulatedModel::Preprocess preprocess;
    preprocess.xform = Matrix4::scale(scale);
    preprocess.replaceTwoSidedWithGeometry = true;

    ArticulatedModel::Ref m = ArticulatedModel::fromFile(filename, preprocess);
    m->pose(m_model);

    // Extract all geometry by loading again and merging everything.  This
    // is inefficient because we parse everything twice...but it is easy.
    preprocess.stripMaterials = true;
    ArticulatedModel::MergeOperation::Ref merge = ArticulatedModel::MergeOperation::create();
    merge->part.append(ArticulatedModel::PartID(ArticulatedModel::ALL));
    preprocess.program.append(merge);
    ArticulatedModel::Ref m2 = ArticulatedModel::fromFile(filename, preprocess);

//    alwaysAssertM(m2->partArray.size() == 1, "Too many parts");
//    alwaysAssertM(m2->partArray[0].triList.size() == 1, "Too many trilists");

    const ArticulatedModel::Part& part = m2->partArray[0];
    m_geometry = IFSModel::fromData(filename, part.geometry.vertexArray, part.indexArray, Array<Vector2>(), Vector3(1,1,1), part.cframe, true, true)->pose();

    defaultCamera.setCoordinateFrame(CFrame::fromXYZYPRDegrees( -0.5f,  -0.4f,   1.2f,  322.5f,  18.8f,   0.0f));
    defaultCamera.setNearPlaneZ(-0.2f);
    defaultCamera.setFarPlaneZ(-40.0f);
}


void App::loadShaders() {
    m_depthShader = Shader::fromFiles("", "shader/depth.pix");
    m_normalShader = Shader::fromFiles("shader/normal.vrt", "shader/normal.pix");
}


void App::makeGui() {
    debugWindow->setVisible(true);
    developerWindow->cameraControlWindow->setVisible(false);
    developerWindow->videoRecordDialog->setEnabled(true);
    showRenderingStats = false;

    debugWindow->setCaption("Style");
    GuiPane* pane = debugWindow->pane();
    pane->addLabel("Style:");
    pane->addRadioButton("None", NO_STYLE, &m_style);   
    pane->addRadioButton("Depth Buffer", DEPTH_STYLE, &m_style);
    pane->addNumberBox("Threshold", &m_depthThreshold, "", GuiTheme::LOG_SLIDER, 0.0001f, 0.01f)->moveBy(20, 0);

    pane->addRadioButton("N dot E", NORMAL_STYLE, &m_style);
    pane->addNumberBox("Threshold", &m_normalThreshold, "", GuiTheme::LINEAR_SLIDER, 0.1f, 1.0f)->moveBy(20, 0);

    pane->addRadioButton("Geometric", GEOMETRIC_STYLE, &m_style);
    
    GuiTextureBox* t = pane->addTextureBox("Normal Buffer", m_gbuffer->csNormal(), GuiTextureBox::Settings::unitVector());
    pane->addTextureBox("Depth Buffer", m_gbuffer->depth(), GuiTextureBox::Settings::depthBuffer())->moveRightOf(t);
    pane->pack();
}

void App::onPose(Array<Surface::Ref>& surfaceArray, Array<Surface2D::Ref>& surface2D) {
    // Append any models to the arrays that you want to later be rendered by onGraphics()
    surfaceArray.append(m_model);
    (void)surface2D;
}


void App::onGraphics3D(RenderDevice* rd, Array<Surface::Ref>& surface3D) {
    m_gbuffer->compute(rd, defaultCamera, m_model);

    if (m_lighting->environmentMap.notNull()) {
        Draw::skyBox(rd, m_lighting->environmentMap);
    }

    // Set a polygon offset so that lines are slightly in front of the surface
    rd->setPolygonOffset(-0.5f);

    Surface::sortAndRender(rd, defaultCamera, surface3D, m_lighting, m_shadowMap);

    switch (m_style) {
    case NO_STYLE:
        break;

    case DEPTH_STYLE:
        rd->push2D(); {
            m_depthShader->args.set("depthBuffer", m_gbuffer->depth());
            m_depthShader->args.set("threshold", m_depthThreshold);
            rd->setShader(m_depthShader);
            rd->setBlendFunc(RenderDevice::BLEND_ZERO, RenderDevice::BLEND_SRC_COLOR);

            Draw::fastRect2D(rd->viewport(), rd);        
        } rd->pop2D();
        break;

    case NORMAL_STYLE:
        rd->pushState(); {
            m_normalShader->args.set("threshold", m_normalThreshold);
            rd->setShader(m_normalShader);
            rd->setBlendFunc(RenderDevice::BLEND_ZERO, RenderDevice::BLEND_SRC_COLOR);

            const CFrame& cameraFrameInverse = defaultCamera.coordinateFrame().inverse();

            for (int i = 0; i < m_model.size(); ++i) {
                const Surface::Ref s = m_model[i];
                const CFrame& modelFrame = s->coordinateFrame();

                const CFrame& o2c = cameraFrameInverse * modelFrame;
                rd->setObjectToWorldMatrix(modelFrame);
                m_normalShader->args.set("normalMatrix", o2c.rotation);
                m_normalShader->args.set("pointMatrix", Matrix4(o2c));
                m_normalShader->args.set("mvpMatrix", rd->modelViewProjectionMatrix());
                s->sendGeometry(rd);
            }
        } rd->popState();
        break;

    case GEOMETRIC_STYLE:
        rd->pushState(); {
            rd->setLineWidth(5);
            rd->setColor(Color3::black());
            drawFeatureEdges(rd, m_geometry, pif() / 2.0f);
        } rd->popState();
        break;

    }

    // Call to make the GApp show the output of debugDraw
    drawDebugShapes();
}
